import threading
import time

from . import _impl
from .common import *
from .connection import *
from .networktablenode import NetworkTableNode
from .type import NetworkTableEntryTypeManager

import logging
logger = logging.getLogger('nt')

__all__ = ["NetworkTableServer"]

class ServerConnectionState:
    """Represents the state of a connection to the server
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

# represents that the server has received the connection from the client but
# has not yet received the client hello
GOT_CONNECTION_FROM_CLIENT = ServerConnectionState("GOT_CONNECTION_FROM_CLIENT")
# represents that the client is in a connected non-error state
CONNECTED_TO_CLIENT = ServerConnectionState("CONNECTED_TO_CLIENT")
# represents that the client has disconnected from the server
CLIENT_DISCONNECTED = ServerConnectionState("CLIENT_DISCONNECTED")

class ServerError(ServerConnectionState):
    """Represents that the client is in an error state
    """

    def __init__(self, e):
        """Create a new error state
        :param e:
        """
        ServerConnectionState.__init__(self, "SERVER_ERROR")
        self.e = e

    def getException(self):
        """:returns: the exception that caused the client connection to enter
            an error state
        """
        return self.e

    def __str__(self):
        return "SERVER_ERROR: %s" % self.e

class ServerConnectionAdapter:
    """Object that adapts messages from a client to the server
    """

    def gotoState(self, newState):
        if self.connectionState != newState:
            logger.info("%s entered connection state: %s", self, newState)
            self.connectionState = newState

    def __init__(self, stream, entryStore, adapterListener, typeManager):
        """Create a server connection adapter for a given stream

        :param stream:
        :param entryStore:
        :param adapterListener:
        """
        self.connection = NetworkTableConnection(stream, typeManager)
        self.entryStore = entryStore
        self.adapterListener = adapterListener

        self.connectionState = None
        self.gotoState(GOT_CONNECTION_FROM_CLIENT)
        self.readManager = ReadManager(self,
                self.connection, name="Server Connection Reader Thread")
        self.readManager.start()
        
    def __str__(self):
        return 'Server 0x%08x' % id(self)

    def badMessage(self, e):
        self.gotoState(ServerError(e))
        self.adapterListener.close(self, True)

    def ioError(self, e):
        if isinstance(e, StreamEOF):
            self.gotoState(CLIENT_DISCONNECTED)
        else:
            self.gotoState(ServerError(e))
        self.adapterListener.close(self, False)

    def shutdown(self, closeStream):
        """stop the read thread and close the stream
        """
        self.readManager.stop()
        if closeStream:
            self.connection.close()

    def keepAlive(self):
        pass # just let it happen

    def clientHello(self, protocolRevision):
        if self.connectionState != GOT_CONNECTION_FROM_CLIENT:
            raise BadMessageError("A server should not receive a client hello after it has already connected/entered an error state")
        if protocolRevision != PROTOCOL_REVISION:
            self.connection.sendProtocolVersionUnsupported()
            raise BadMessageError("Client Connected with bad protocol revision: 0x%x" % protocolRevision)
        else:
            self.entryStore.sendServerHello(self.connection)
            self.gotoState(CONNECTED_TO_CLIENT)

    def protocolVersionUnsupported(self, protocolRevision):
        raise BadMessageError("A server should not receive a protocol version unsupported message")

    def serverHelloComplete(self):
        raise BadMessageError("A server should not receive a server hello complete message")

    def offerIncomingAssignment(self, entry):
        self.entryStore.offerIncomingAssignment(entry)

    def offerIncomingUpdate(self, entry, sequenceNumber, value):
        self.entryStore.offerIncomingUpdate(entry, sequenceNumber, value)

    def getEntry(self, id):
        return self.entryStore.getEntry(id)

    def sendEntry(self, entryBytes):
        try:
            if self.connectionState == CONNECTED_TO_CLIENT:
                self.connection.sendEntry(entryBytes)
        except IOError as e:
            self.ioError(e)
    
    def flush(self):
        try:
            self.connection.flush()
        except IOError as e:
            self.ioError(e)

    def getConnectionState(self):
        """:returns: the state of the connection
        """
        return self.connectionState

    def ensureAlive(self):
        try:
            self.connection.sendKeepAlive()
        except IOError as e:
            self.ioError(e)

class ServerNetworkTableEntryStore(AbstractNetworkTableEntryStore):
    """The entry store for a {@link NetworkTableServer}
    """

    def __init__(self, listenerManager):
        """Create a new Server entry store
        :param listenerManager: the listener manager that fires events from
            this entry store
        """
        AbstractNetworkTableEntryStore.__init__(self, listenerManager)
        self.nextId = 0

    def addEntry(self, newEntry):
        with self.entry_lock:
            entry = self.namedEntries.get(newEntry.name)

            if entry is None:
                newEntry.setId(self.nextId)
                self.nextId += 1
                self.idEntries[newEntry.getId()] = newEntry
                self.namedEntries[newEntry.name] = newEntry
                return True
            return False

    def updateEntry(self, entry, sequenceNumber, value):
        with self.entry_lock:
            if entry.putValue(sequenceNumber, value):
                return True
            return False

    def sendServerHello(self, connection):
        """Send all entries in the entry store as entry assignments in a
        single transaction
        :param connection:
        """
        transaction = []
        with self.entry_lock:
            # Cannot use sendEntry while holding entry lock!
            for entry in self.namedEntries.values():
                transaction.append(entry.getAssignmentBytes())
                
        for entry in transaction:
            connection.sendEntry(entry)
        connection.sendServerHelloComplete()
        connection.flush()

class ServerConnectionList:
    """A list of connections that the server currently has
    """

    def __init__(self):
        self.connections = []
        self.connectionsLock = _impl.create_rlock('server_conn_lock')

    def add(self, connection):
        """Add a connection to the list
        :param connection:
        """
        with self.connectionsLock:
            self.connections.append(connection)

    def close(self, connectionAdapter, closeStream):
        with self.connectionsLock:
            try:
                self.connections.remove(connectionAdapter)
            except ValueError:
                return
            logger.info("Close: %s", connectionAdapter)
            connectionAdapter.shutdown(closeStream)

    def closeAll(self):
        """close all connections and remove them
        """
        with self.connectionsLock:
            for connection in self.connections:
                logger.info("Close: %s", connection)
                connection.shutdown(True)
            del self.connections[:]

    def sendEntry(self, entryBytes):
        with self.connectionsLock:
            for connection in self.connections:
                connection.sendEntry(entryBytes)

    def flush(self):
        with self.connectionsLock:
            for connection in self.connections:
                connection.flush()

    def ensureAlive(self):
        with self.connectionsLock:
            for connection in self.connections:
                connection.ensureAlive()

class NetworkTableServer(NetworkTableNode):
    """A server node in NetworkTables 2.0
    """

    def __init__(self, streamProvider):
        """Create a NetworkTable Server
        :param streamProvider:
        """
        NetworkTableNode.__init__(self, ServerNetworkTableEntryStore(self))
        self.typeManager = NetworkTableEntryTypeManager()
        self.streamProvider = streamProvider

        self.connectionList = ServerConnectionList()
        self.writeManager = WriteManager(self.connectionList, self.entryStore, None)

        self.entryStore.setIncomingReceiver(self.writeManager)
        self.entryStore.setOutgoingReceiver(self.writeManager)

        # start incoming stream monitor
        self.running = True
        self.monitorThread = threading.Thread(target=self._incomingMonitor,
                                              name="Server Incoming Stream Monitor Thread")
        self.monitorThread.daemon = True
        self.monitorThread.start()

        # start write manager
        self.writeManager.start()

    def close(self):
        try:
            self.running = False
            self.monitorThread.join()
            self.writeManager.stop()
            self.connectionList.closeAll()
            time.sleep(1) #To get around bug where an error will occur in select if the socket server is closed before all sockets finish closing
            self.streamProvider.close()
            time.sleep(1)
        except IOError as e:
            logger.error("Error during close: %s", e)

    def isConnected(self):
        return True

    def isServer(self):
        return True

    def _incomingMonitor(self):
        while self.running:
            try:
                newStream = self.streamProvider.accept()
                if newStream is not None:
                    connectionAdapter = ServerConnectionAdapter(newStream, self.entryStore, self.connectionList, self.typeManager)
                    self.connectionList.add(connectionAdapter)
            except IOError:
                pass #could not get a new stream for some reason. ignore and continue




# Common classes for client and server

import threading
import time
import warnings

from . import _impl
from .entry import NetworkTableEntry

__all__ = ["AbstractNetworkTableEntryStore", "WriteManager"]

class AbstractNetworkTableEntryStore:
    """An entry store that handles storing entries and applying transactions
    """

    def __init__(self, listenerManager):
        self.idEntries = {}
        self.namedEntries = {}
        self.listenerManager = listenerManager
        self.outgoingReceiver = None
        self.incomingReceiver = None
        self.entry_lock = _impl.create_rlock('entry_lock')

    def getEntry(self, name_id):
        """Get an entry based on its name or id
        :param name_id: the name or id of the entry to look for
        :returns: the entry or None if the entry does not exist
        """
        with self.entry_lock:
            if isinstance(name_id, str):
                return self.namedEntries.get(name_id)
            else:
                return self.idEntries.get(name_id)

    def keys(self):
        """Get the list of keys.
        :returns: the list of keys
        """
        with self.entry_lock:
            return [k for k in self.namedEntries.keys()]

    def clearEntries(self):
        """Remove all entries.
        NOTE: This method should not be used with applications which cache
        entries which would lead to unknown results.
        This method is for use in testing only.
        """
        with self.entry_lock:
            self.idEntries.clear()
            self.namedEntries.clear()

    def clearIds(self):
        """clear the id's of all entries
        """
        with self.entry_lock:
            self.idEntries.clear()
            for entry in self.namedEntries.values():
                entry.clearId()

    def setOutgoingReceiver(self, receiver):
        self.outgoingReceiver = receiver

    def setIncomingReceiver(self, receiver):
        self.incomingReceiver = receiver

    def addEntry(self, entry):
        raise NotImplementedError

    def updateEntry(self, entry, sequenceNumber, value):
        raise NotImplementedError

    def putOutgoing(self, name, type, value):
        """Stores the given value under the given name and queues it for
        transmission to the server.

        :param name: The name under which to store the given value.
        :param type: The type of the given value.
        :param value: The value to store.
        Raises TypeError if entry already exists with the given name and is of
        a different type.
        """
        with self.entry_lock:
            tableEntry = self.namedEntries.get(name)
            if tableEntry is None:
                #TODO validate type
                tableEntry = NetworkTableEntry(name, type, value)
                if self.addEntry(tableEntry):
                    tableEntry.fireListener(self.listenerManager)
                    if self.outgoingReceiver is not None:
                        self.outgoingReceiver.offerOutgoingAssignment(tableEntry)
            else:
                if tableEntry.getType().id != type.id:
                    raise TypeError(name)
                if value != tableEntry.getValue():
                    if self.updateEntry(tableEntry, tableEntry.getSequenceNumber()+1, value):
                        if self.outgoingReceiver is not None:
                            self.outgoingReceiver.offerOutgoingUpdate(tableEntry)
                    tableEntry.fireListener(self.listenerManager)

    def putOutgoingEntry(self, tableEntry, value):
        with self.entry_lock:
            #TODO Validate type
            if value != tableEntry.getValue():
                if self.updateEntry(tableEntry, tableEntry.getSequenceNumber()+1, value):
                    if self.outgoingReceiver is not None:
                        self.outgoingReceiver.offerOutgoingUpdate(tableEntry)
                tableEntry.fireListener(self.listenerManager)

    def offerIncomingAssignment(self, entry):
        with self.entry_lock:
            tableEntry = self.namedEntries.get(entry.name)
            if self.addEntry(entry):
                if tableEntry is None:
                    tableEntry = entry
                tableEntry.fireListener(self.listenerManager)
                if self.incomingReceiver is not None:
                    self.incomingReceiver.offerOutgoingAssignment(tableEntry)

    def offerIncomingUpdate(self, entry, sequenceNumber, value):
        with self.entry_lock:
            if self.updateEntry(entry, sequenceNumber, value):
                entry.fireListener(self.listenerManager)
                if self.incomingReceiver is not None:
                    self.incomingReceiver.offerOutgoingUpdate(entry)

    def notifyEntries(self, table, listener):
        """Called to say that a listener should notify the listener manager
        of all of the entries
        :param listener:
        :param table:
        """
        with self.entry_lock:
            for entry in self.namedEntries.values():
                listener.valueChanged(table, entry.name, entry.getValue(), True)

class WriteManager:
    """A write manager is a IncomingEntryReceiver that buffers transactions
    and then dispatches them to a flushable transaction receiver that is
    periodically offered all queued transaction and then flushed
    """
    SLEEP_TIME = 0.050
    
    queueSize = 500

    def __init__(self, receiver, entryStore, keepAliveDelay):
        """Create a new Write manager
        :param receiver:
        :type receiver: :class:`.ServerConnectionList`, :class:`.ClientConnectionAdapter`
        :param entryStore:
        """
        self.receiver = receiver
        self.entryStore = entryStore
        self.keepAliveDelay = keepAliveDelay
        self.lastWrite = 0

        self.transactionsLock = _impl.create_rlock('trans_lock')
        self.transactionsCondition = threading.Condition(self.transactionsLock)

        self.incomingAssignmentQueue = []
        self.incomingUpdateQueue = []
        self.outgoingAssignmentQueue = []
        self.outgoingUpdateQueue = []

        self.thread = None
        self.running = False

    def start(self):
        """start the write thread
        """
        if self.thread is not None:
            self.stop()
        self.lastWrite = time.time()
        self.running = True
        self.thread = threading.Thread(target=self.run,
                                       name="Write Manager Thread")
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        """stop the write thread
        """
        if self.thread is not None:
            self.running = False
            with self.transactionsCondition:
                self.transactionsCondition.notify()
            self.thread.join()

    def offerOutgoingAssignment(self, entry):
        # This is always called with the entry lock held
        
        # Mark entry as dirty to avoid duplicate updates
        if entry.isDirty:
            return
        entry.makeDirty()

        with self.transactionsLock:
            self.incomingAssignmentQueue.append(entry)
            if len(self.incomingAssignmentQueue) >= self.queueSize:
                warnings.warn("assignment queue overflowed. decrease the rate at which you create new entries or increase the write buffer size", ResourceWarning)
                self.transactionsCondition.notify()

    def offerOutgoingUpdate(self, entry):
        # This is always called with the entry lock held
        
        # Mark entry as dirty to avoid duplicate updates
        if entry.isDirty:
            return
        entry.makeDirty()

        with self.transactionsLock:
            self.incomingUpdateQueue.append(entry)
            if len(self.incomingUpdateQueue) >= self.queueSize:
                warnings.warn("update queue overflowed. decrease the rate at which you update entries or increase the write buffer size", ResourceWarning)
                self.transactionsCondition.notify()

    def run(self):
        """the periodic method that sends all buffered transactions
        """
        while self.running:
            
            with self.transactionsLock:
                
                self.transactionsCondition.wait(self.SLEEP_TIME)
                
                if not self.running:
                    break
                
                #swap the assignment and update queue
                self.incomingAssignmentQueue, self.outgoingAssignmentQueue = \
                    self.outgoingAssignmentQueue, self.incomingAssignmentQueue
    
                self.incomingUpdateQueue, self.outgoingUpdateQueue = \
                    self.outgoingUpdateQueue, self.incomingUpdateQueue
            
            # Decision: Choose to lock/unlock the entry lock quickly, instead
            #           of one big lock. This allows the main thread to not
            #           be interrupted for an extended period of time
            
            transactions = []
    
            for entry in self.outgoingAssignmentQueue:
                with self.entryStore.entry_lock:
                    entry.makeClean()
                    transactions.append(entry.getAssignmentBytes())
                    
            for entry in self.outgoingUpdateQueue:
                with self.entryStore.entry_lock:
                    entry.makeClean()
                    transactions.append(entry.getUpdateBytes())
                
            del self.outgoingAssignmentQueue[:]
            del self.outgoingUpdateQueue[:]
    
            for entry in transactions:
                self.receiver.sendEntry(entry)
    
            if len(transactions) > 0:
                self.receiver.flush()
                self.lastWrite = time.time()
            elif (self.keepAliveDelay is not None and
                  (time.time()-self.lastWrite) > self.keepAliveDelay):
                self.receiver.ensureAlive()
